/*****************************************************************************
 * va.c: Video Acceleration API for avcodec
 *****************************************************************************
 * Copyright (C) 2011, M2X BV
 * $Id: 87c0614ce43b4ba7e0c58710e05c7cdeeffbb24e $
 *
 * Authors: Jean-Paul Saman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_fourcc.h>
#include <assert.h>

#ifdef HAVE_LIBAVCODEC_AVCODEC_H
#   include <libavcodec/avcodec.h>
#else
#   include <avcodec.h>
#endif

#include "avcodec.h"
#ifdef HAVE_AVCODEC_VAAPI
#  include <vlc_xlib.h>
#  include <libavcodec/vaapi.h>
#  include <va/va.h>
#  include <va/va_x11.h>
#endif
#include "va.h"
#include "vaapi.h"

#ifdef HAVE_AVCODEC_VAAPI

static vlc_mutex_t vlc_va_conn_lock = VLC_STATIC_MUTEX;

static void vlc_va_lock(void)
{
    vlc_mutex_lock(&vlc_va_conn_lock);
}

static void vlc_va_unlock(void)
{
    vlc_mutex_unlock(&vlc_va_conn_lock);
}

/* */
static VASurfaceID *vlc_va_alloc_surfaces( vlc_va_conn_t *conn,
    const int width, const int height, const int fourcc, unsigned int *requested )
{
    assert( *requested > 0 );

    VASurfaceID *p_surfaces = (VASurfaceID *)calloc(*requested, sizeof(VASurfaceID));
    if( p_surfaces == NULL )
        return NULL;

    VAStatus status = vaCreateSurfaces( conn->p_display, width, height, fourcc,
                                        *requested, p_surfaces );
    if( status != VA_STATUS_SUCCESS )
    {
        free( p_surfaces );
        return NULL;
    }

    return p_surfaces;
}

static VASurfaceID *vlc_va_realloc_surfaces( vlc_va_conn_t *conn,
    const int width, const int height, const int fourcc, unsigned int *requested )
{
    assert( *requested > 0 );

    VASurfaceID *p_surfaces = (VASurfaceID *)realloc( conn->p_surface_ids,
                                                     (*requested) * sizeof(VASurfaceID) );
    if( p_surfaces == NULL )
        return NULL;

    int count = *requested - conn->i_surface_count;
    assert( count > 0 );

    VASurfaceID *p_tmp = &p_surfaces[conn->i_surface_count + 1];
    VAStatus status = vaCreateSurfaces( conn->p_display, width, height, fourcc,
                                        count, p_tmp );
    if( status != VA_STATUS_SUCCESS )
    {
        free( p_surfaces );
        return NULL;
    }
    return p_surfaces;
}

static VASurfaceID *vlc_va_create_surfaces( vlc_va_conn_t *conn,
    const int width, const int height, const int fourcc, const unsigned int num )
{
    assert(conn);

    if( (width <= 0) || (height <= 0) || (num <= 0) )
        return NULL;

    if( conn->i_ref_count > 1 )
    {
        assert( conn->p_surface_ids != NULL );

        if( num > conn->i_surface_count )
        {
            assert(0);
            /* FIXME: probably wise to destroy decoding context first
               and then recreate it */
            unsigned requested = num;
            vlc_va_realloc_surfaces( conn, width, height, fourcc, &requested );
            assert( num == requested );
        }
        return conn->p_surface_ids;
    }

    assert( conn->p_surface_ids == NULL );
    assert( conn->i_surface_count == 0 );

    unsigned requested = num;
    conn->p_surface_ids = vlc_va_alloc_surfaces( conn, width, height,
                                                fourcc, &requested );
    assert( num == requested );
    if( conn->p_surface_ids )
        conn->i_surface_count = requested;
    return conn->p_surface_ids;
}

static void vlc_va_destroy_surfaces( vlc_va_conn_t *conn )
{
    assert( conn );
    assert( conn->i_surface_count > 0 );
    assert( conn->p_surface_ids );

    if( conn->i_ref_count > 1 )
        return;

    for( unsigned int i = 0; i < conn->i_surface_count; i++ )
    {
        VASurfaceID i_id = conn->p_surface_ids[i];
#if VA_CHECK_VERSION(0,31,0)
        if( vaSyncSurface( conn->p_display, i_id ) )
#else
        if( vaSyncSurface( conn->p_display, conn->i_context_id, i_id ) )
#endif
        {
            fprintf(stderr, "Fatal error reboot your system\n" );
            abort();
        }
    }

    if( conn->p_surface_ids != NULL )
        vaDestroySurfaces( conn->p_display, conn->p_surface_ids,
                           conn->i_surface_count );

    free(conn->p_surface_ids );

    conn->i_surface_count = 0;
    conn->p_surface_ids = NULL;
}

/* Global VAAPI connection state */
static vlc_va_conn_t vlc_va_conn = {
    .x11 = NULL,
    .p_display = 0,
    /* libva version */
    .i_version_major = 0,
    .i_version_minor = 0,
    /* ref counting */
    .i_ref_count = 0,
    /* locking functions */
    .lock = vlc_va_lock,
    .unlock = vlc_va_unlock,
    /* surfaces - NOTE: must be called with lock held */
    .i_surface_count = 0,
    .p_surface_ids = NULL,
    .i_context_id = VA_INVALID_ID,
    .create_surfaces = vlc_va_create_surfaces,
    .destroy_surfaces = vlc_va_destroy_surfaces,
};

/* */
static vlc_va_conn_t *vlc_va_get_conn( void )
{
    return (vlc_va_conn_t *) &vlc_va_conn;
}

vlc_va_conn_t *vlc_va_Initialize( const char *display )
{
    vlc_mutex_lock( &vlc_va_conn_lock );

    /* connect global to caller */
    vlc_va_conn_t *conn = vlc_va_get_conn();
    assert( conn );
    if( conn->i_ref_count > 0 )
    {
        conn->i_ref_count++;
        vlc_mutex_unlock( &vlc_va_conn_lock );
        return conn;
    }

    /* X11 display */
    conn->x11 = XOpenDisplay(display);
    if( !conn->x11 )
        goto error;

    /* Create a VA display */
    conn->p_display = vaGetDisplay(conn->x11);
    if( !conn->p_display )
        goto error;

    if( vaInitialize( conn->p_display, &conn->i_version_major, &conn->i_version_minor ) )
        goto error;

    conn->i_ref_count++;
    vlc_mutex_unlock( &vlc_va_conn_lock );
    return conn;

error:
    if( conn->x11 )
        XCloseDisplay( conn->x11 );
    vlc_mutex_unlock( &vlc_va_conn_lock );
    return NULL;
}

void vlc_va_Terminate( vlc_va_conn_t *conn )
{
    vlc_mutex_lock( &vlc_va_conn_lock );
    assert( conn );
    assert( conn->i_ref_count > 0 );

    conn->i_ref_count--;
    if( conn->i_ref_count == 0 )
    {
        assert( conn->i_surface_count == 0 );
        assert( conn->p_surface_ids == NULL );

        assert( conn->p_display );
        vaTerminate( conn->p_display );
        XCloseDisplay( conn->x11 );

        /* Reset values */
        conn->x11 = NULL;
        conn->p_display = 0;
        conn->i_version_major = conn->i_version_minor = 0;
        conn->i_ref_count = 0;
        conn = NULL;
    }

    vlc_mutex_unlock( &vlc_va_conn_lock );
}

#endif
